from .path_alg import solve_path
import numpy as np
import numpy.linalg as LA
from .solve_R1 import problem_R1, Classo_R1
    
'''    
Problem    :   min h_rho(Ab - y) + lambda ||b||1 with C.b= 0 <=>   min ||Ab - y - r*o||^2 + lambda ||b,o||1 with C.b= 0, o in R^m
                                                                                        r = lambda / 2rho
Dimensions :   A : m*d  ;  y : m  ;  b : d   ; C : k*d

The first function compute a solution of a Lasso problem for a given lambda. The parameters are lam (lambda/lambdamax, \in [0,1]) and pb, which has to be a 'problem_LS type', which is defined bellow in order to contain all the important parameters of the problem. One can initialise it this way : pb = class_of_problem.problem(data=(A,C,y),type_of_algo). We solve the problem without normalizing anything. 
'''    


def Classo_R2(pb,lam, compute=True):
    
    pb_type = pb.type    # can be 'Path-Alg', 'P-PDS' , 'PF-PDS' or 'DR'
    
    
    (m,d,k),(A,C,y)  = pb.dim,pb.matrix        
    lamb, rho = lam * pb.lambdamax,  pb.rho
    
    
    #ODE
    # here we compute the path algo until our lambda, and just take the last beta
    
    if(pb_type == 'Path-Alg'):
        beta = solve_path((A,C,y), lam,False,rho,'huber')[0]
        return(beta[-1])

    # 2 prox :
    regpath = pb.regpath
    r = lamb / (2 * rho)
    if (pb_type == 'DR'):
        Ahuber = np.concatenate((A, r * np.eye(len(A))), axis=1)
        Chuber = np.concatenate((C, np.zeros((len(C), len(y)))), axis=1)
        matrices_huber = (Ahuber, Chuber, y)
        prob = problem_R1(matrices_huber, 'DR')
        prob.regpath = regpath
        if (len(pb.init) == 3): prob.init = pb.init
        if not (regpath): return (Classo_R1(prob, lamb / prob.lambdamax)[:d])
        x, warm_start = Classo_R1(prob, lamb / prob.lambdamax)
        return (x[:d], warm_start)


    tol = pb.tol * LA.norm(y)/LA.norm(A,'fro')  # tolerance rescaled
    
    #cvx 
    # call to the cvx function of minimization
    if (pb_type == 'cvx'):
        import cvxpy as cp
        x = cp.Variable(d)
        objective, constraints = cp.Minimize(cp.sum(cp.huber(A*x - y, rho))+lamb*cp.norm(x, 1)), [C*x == 0]
        prob = cp.Problem(objective, constraints)
        result = prob.solve(warm_start=regpath,eps_abs= tol)
        if (regpath): return(x.value,True) 
        return(x.value)
     
    if(compute): pb.compute_param()
    tau,Proj, AtA, Aty = pb.tauN,  proj_c(C,d), pb.AtA, pb.Aty
    gamma                 = pb.gam / (2*(pb.AtAnorm+r**2))   
    t                     = lamb * gamma
    w,tm,zerom,zerod      = t*pb.weights,t*np.ones(m), np.zeros(m), np.zeros(d) 
    o,xbar,x,v = pb.init
    # vectors usefull to compute the prox of f(b)= sum(wi |bi|)
    
    #FORWARD BACKWARD
    if (pb_type=='P-PDS'):
        
        for i in range(pb.N):                        
            grad = (AtA.dot(x)-Aty)                  
            v = v + tau*C.dot(xbar)             
            S = x - 2*gamma*grad - 2*gamma*r*(A.T).dot(o) - (C.T).dot(v) 
            o = prox(o*(1-2*gamma*r**2) + 2*gamma*r*(y-A.dot(x)),tm,zerom)
            p = prox(S,w,zerod)             
            nw_x = Proj.dot(p)                   
            eps = nw_x - x                      
            xbar= p + eps                     

            if (i%10==2 and LA.norm(eps)<tol):      # 0.6
                if (regpath): return(x,(o,xbar,x,v)) 
                else :        return(x) 
            x= nw_x 
            if (LA.norm(x)>1e10): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
     
        
    # NO PROJ
    
    if (pb_type == 'PF-PDS'):     # y1 --> S ; p1 --> p . ; p2 --> y2
        for i in range(pb.N):    
            grad = (AtA.dot(x)-Aty)
            
            S1 = x - 2*gamma*grad - 2*gamma*r*(A.T).dot(o) - (C.T).dot(v)
            S2 = o*(1-2*gamma*r**2) + 2*gamma*r*(y-A.dot(x))
            
            p1 = prox(S1,w,zerod)
            p2 = prox(S2,tm,zerom)
            
            v   = v + tau*C.dot(p1)
            v2  = v + tau*C.dot(x)
            
            eps1 = p1 + 2*gamma*(Aty-AtA.dot(p1)-r*A.T.dot(o)) - C.T.dot(v2)-S1
            eps2 = p2 + 2*r*gamma*(y-r*p2-A.dot(x)) -S2
            
            x = x + eps1
            o = o + eps2
            
            if (i>0 and LA.norm(eps1)+LA.norm(eps2)<tol): 
                if (regpath): return(x,(o,xbar,x,v)) 
                else :        return(x) 
            
            if (LA.norm(x)+LA.norm(o)+LA.norm(v)>1e6): 
                print('DIVERGENCE')
                return(x)
        print('NO CONVERGENCE')
        return(x)
    
    
    
    

    
'''
This function compute the the solution for a given path of lam : by calling the function 'algo' for each lambda with warm start, or wuth the method ODE, by computing the whole path thanks to the ODE that rules Beta and the subgradient s, and then to evaluate it in the given finite path.  
'''
    
    
def pathlasso_R2(pb,path,n_active=False):
    n = pb.dim[0]
    BETA,tol = [],pb.tol
    if(pb.type == 'Path-Alg'):
        X,sp_path = solve_path(pb.matrix,path[-1],n_active,pb.rho, 'huber')
        # we do a little manipulation to interpolated the value of beta between breaking points, as we know beta is affine between those those points.
        i=0
        sp_path.append(path[-1]),X.append(X[-1])
        for lam in path:
            while (lam<sp_path[i+1]): i+=1
            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
            BETA.append(X[i]*(1-teta)+X[i+1]*teta)
        return(BETA)

    # Now we are in the case where we have to do warm starts.
    save_init = pb.init   
    pb.regpath = True
    pb.compute_param()
    if (type(n_active)==int) : n_act = n_active
    else : n_act = n
    for lam in path:
        X = Classo_R2(pb,lam,compute=False)
        BETA.append(X[0])
        pb.init = X[1]
        if(sum([ (abs(X[0][i])>1e-2) for i in range(len(X[0])) ])>=n_act):
                pb.init = save_init
                BETA = BETA + [BETA[-1]]*(len(path)-len(BETA))
                pb.regpath = False
                return(BETA)
            
    pb.init = save_init
    pb.regpath = False
    return(BETA)








'''
Class of problem : we define a type, which will contain as keys, all the parameters we need for a given problem.
'''


class problem_R2 :
    
    def __init__(self,data,algo,rho):
        self.N = 500000
        
        if(len(data)==3):self.matrix, self.dim = data, (data[0].shape[0],data[0].shape[1],data[1].shape[0])
        
        elif(len(data)==5):
            (A,C,sol), self.dim = generate_random(data), data[:3]
            self.sol,y = sol, A.dot(sol)+np.random.randn(data[0])*data[-1]
            self.matrix = (A,C,y)
        
        (m,d,k) = self.dim
        self.weights = np.ones(d)
        self.init = np.zeros(m), np.zeros(d), np.zeros(d), np.zeros(k)
        self.tol = 1e-6
        self.regpath = False
        self.name = algo + ' Huber'
        self.type = algo        # type of algo used
        self.rho = rho
        self.gam = 1.
        self.tau = 0.5         # equation for the convergence of Noproj and LS algorithms : gam + tau < 1
        self.lambdamax = 2*LA.norm(self.matrix[0].T.dot(h_prime(self.matrix[2],rho)),np.infty)


    '''
    this is a method of the class pb that is used to computed the expensive multiplications only once. (espacially usefull for warm start. )
    '''
    def compute_param(self):
        (A,C,y) = self.matrix
        m,d,k = self.dim
        self.c = (d/LA.norm(A,2))**2  # parameter for Concomitant problem : the matrix is scaled as c*A^2

        self.AtA        =(A.T).dot(A)
        self.Aty        = (A.T).dot(y)
        self.Cnorm      = LA.norm(C,2)**2
        self.tauN       = self.tau/self.Cnorm
        self.AtAnorm    = LA.norm(self.AtA,2)
        


# compute the prox of the function : f(b)= sum (wi * |bi| )
def prox(b,w,zeros): return(np.minimum(b+w,zeros)+np.maximum(b-w,zeros)) 

# Compute I - C^t (C.C^t)^-1 . C : the projection on Ker(C)
def proj_c(M,d):
    if (LA.matrix_rank(M)==0):  return(np.eye(d))
    return(np.eye(d)-LA.multi_dot([M.T,np.linalg.inv(M.dot(M.T) ),M]) )

# Compute the derivative of the huber function, particulary useful for the computing of lambdamax 
def h_prime(y,rho):
    m = len(y)
    lrho = rho*np.ones(m)
    return(np.maximum(lrho,-y)+ np.minimum(y-lrho,0))





import numpy as np
import numpy.linalg as LA

from .solve_R1 import problem_R1, Classo_R1, pathlasso_R1
from .solve_R2 import problem_R2, Classo_R2, pathlasso_R2
from .solve_R3 import problem_R3, Classo_R3, pathlasso_R3
from .solve_R4 import problem_R4, Classo_R4, pathlasso_R4
from .path_alg import solve_path, pathalgo_general, h_lambdamax

'''
Classo and pathlasso are the main functions, they can call every algorithm acording to the method and formulation required
'''

# can be 'Path-Alg', 'P-PDS' , 'PF-PDS' or 'DR'

def Classo(matrix,lam,typ = 'R1', meth='DR', rho = 1.345, get_lambdamax = False, true_lam=False, e=1., rho_classification=-1.):
    if(typ=='R3'):
        if not meth in ['Path-Alg', 'DR']: meth='DR'
        pb = problem_R3(matrix,meth,e=e)
        if (true_lam): beta,s = Classo_R3(pb,lam/pb.lambdamax)
        else : beta, s = Classo_R3(pb, lam)
        s = s/np.sqrt(e)

    elif(typ=='R4'):
        if not meth in ['Path-Alg', 'DR']: meth='DR'
        pb  = problem_R4(matrix,meth,rho,e=e)
        if (true_lam): beta,s = Classo_R4(pb,lam/pb.lambdamax,e=e)
        else : beta, s = Classo_R4(pb, lam,e=e)


    elif(typ=='R2'):
        if not meth in ['Path-Alg', 'P-PDS' , 'PF-PDS' , 'DR']: meth = 'ODE'
        pb = problem_R2(matrix,meth,rho)
        if (true_lam): beta = Classo_R2(pb,lam/pb.lambdamax)
        else : beta = Classo_R2(pb, lam)

    elif (typ == 'C2'):
        if (true_lam):  BETA = solve_path(matrix, lamin, False, rho_classification, 'huber_cl')[0] #TO DO HERE !!!!!!!!!
        else :    BETA = solve_path(matrix, lamin, False, rho_classification, 'huber_cl')[0]
        beta = BETA[0]

    elif (typ == 'C1'):
        if(true_lam) : BETA = solve_path(matrix,lamin, False,0, 'cl')[0] # TO DO HERE !!!!!!!!
        else : BETA = solve_path(matrix,lamin, False,0, 'cl')[0]
        beta = BETA[0]


    else: # LS
        if not meth in ['Path-Alg', 'P-PDS' , 'PF-PDS' , 'DR']: meth='DR'
        pb = problem_R1(matrix,meth)
        if (true_lam) : beta = Classo_R1(pb,lam/pb.lambdamax)
        else : beta = Classo_R1(pb,lam)

    if (typ  in ['R3','R4']):
        if (get_lambdamax): return(pb.lambdamax,beta,s)
        else              : return(beta,s)
    if (get_lambdamax): return(pb.lambdamax,beta)
    else              : return(beta)


def pathlasso(matrix,lambdas=False,n_active=0,lamin=1e-2,typ='LS',meth='Path-Alg',rho = 1.345, true_lam = False, e= 1.,return_sigm= False,rho_classification=-1):
    Nactive = n_active
    if(Nactive==0):Nactive=False
    if (type(lambdas)!= bool):
        if (lambdas[0]<lambdas[-1]): lambdas = [lambdas[i] for i in range(len(lambdas)-1,-1,-1)]  # reverse the list if needed
    else: lambdas = np.linspace(1.,lamin,100)

    if(typ=='R2'):
        pb = problem_R2(matrix,meth,rho)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        BETA  = pathlasso_R2(pb,lambdas,n_active=Nactive)

    elif(typ=='R3'):
        pb = problem_R3(matrix,meth,e=e)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        BETA,S = pathlasso_R3(pb,lambdas,n_active=Nactive)
        S=np.array(S)/np.sqrt(e)

    elif(typ=='R4'):
        meth='DR'
        pb = problem_R4(matrix,meth,rho)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        BETA,S = pathlasso_R4(pb,lambdas,n_active=Nactive)
        
    elif(typ == 'C2'):
        lambdamax = h_lambdamax(matrix[0],matrix[2],rho)
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        BETA = pathalgo_general(matrix, lambdas, 'huber_cl', n_active=Nactive, rho=rho_classification)

    elif (typ == 'C1'):
        lambdamax = 2*LA.norm((matrix[0].T).dot(matrix[2]),np.infty)
        if (true_lam): lambdas = [lamb / lambdamax for lamb in lambdas]
        BETA = pathalgo_general(matrix, lambdas, 'cl', n_active=Nactive)

    else:
        pb = problem_R1(matrix,meth)
        lambdamax = pb.lambdamax
        if (true_lam): lambdas=[lamb/lambdamax for lamb in lambdas]
        BETA = pathlasso_R1(pb,lambdas,n_active=n_active)

    real_path = [lam*lambdamax for lam in lambdas]
    if(typ in ['R3','R4'] and return_sigm): return(BETA,real_path,S)
    return(BETA,real_path)
 
    


'''
# Cost fucntions for the three 'easiest' problems. Useful for test, to compare two solutions slightly different
def L_LS(A,y,lamb,x): return(LA.norm( A.dot(x) - y )**2 + lamb * LA.norm(x,1))
def L_conc(A,y,lamb,x): return(LA.norm( A.dot(x) - y ) + np.sqrt(2)*lamb * LA.norm(y,1))
def L_H(A,y,lamb,x,rho): return(hub( A.dot(x) - y , rho) + lamb * LA.norm(x,1))

def hub(r,rho) : 
    h=0
    for j in range(len(r)):
        if(abs(r[j])<rho): h+=r[j]**2
        elif(r[j]>0)     : h+= (2*r[j]-rho)*rho
        else             : h+= (-2*r[j]-rho)*rho
    return(h)
'''


N=10000
import numpy as np
import numpy.linalg as LA


class parameters_for_update:
    '''Object that has parameters needed at each breaking point during path algorithm

    Attributes :
        number_act  (int) : number of active parameter
        idr         : saves the independant rows of the matrix C resctricted to the actives parameters
        Xt          :  inverse of M
        activity    :  list of boolean, activity[i] is True when variable i is active
        beta        : current solution beta
        s           : current subgradient
        lam         : current lam
        lambdamax   : lambdamax
        M           : matrix to invert
        y           : output
        r           : residual
        F           : F is the set where r<1 and if huber, then it is the set where rho<r<1
        rho         : only use when doing huber path algo


    '''
    def __init__(self,matrices,lamin, rho,typ):

        (self.A, self.C, self.y) = matrices
        self.lamin = lamin
        self.rho = rho
        self.formulation = typ
        n, d, k = len(self.A), len(self.A[0]), len(self.C)
        self. number_act = 0
        self.idr         = [False] * k
        self.activity    = [False] * d
        self.beta        = np.zeros(d)
        s                = 2*self.A.T.dot(h_prime(self.y,rho))
        self.lambdamax   = LA.norm(s, np.inf)
        self.s           = s/self.lambdamax
        self.lam         = 1.
        self.r           = -self.y
        self.F           = [True] * n
        self.eps_L2      = 1e-3
        if (rho>0):
            for j in range(n):
                if( abs(self.y[j]) > self.rho ): self.F[j] = False
        elif (rho<0):
            for j in range(n):
                if( self.y[j] < self.rho ): self.F[j] = False
        AtA = self.A[self.F].T.dot(self.A[self.F])
        for i in range(d):
            if (self.s[i] == 1. or self.s[i] == -1.):
                self.activity[i] = True
                self.number_act += 1
                if (k > 0):
                    to_ad = next_idr1(self.idr, self.C[:, self.activity])
                    if (type(to_ad) == int): self.idr[to_ad] = True


        if(k==0):  self.M = 2 * AtA
        else    :  self.M  = np.concatenate((np.concatenate((2 * AtA, self.C.T), axis=1), np.concatenate((self.C, np.zeros((k, k))), axis=1)),axis=0)


        try : 
            self.Xt = LA.inv(self.M[self.activity + self.idr, :][:, self.activity + self.idr])
        except LA.LinAlgError :
            self.Xt = LA.inv( self.M[self.activity + self.idr, :][:, self.activity + self.idr] + np.diag([self.eps_L2]*sum(self.activity) + [0]*sum(self.idr) )   )


# iteration of the function up to solve the path at each breaking points.
def solve_path(matrices, lamin, n_active, rho,typ):
    param = parameters_for_update(matrices, lamin, rho,typ)
    BETA, LAM = [param.beta], [param.lam]
    if param.lam < lamin : return BETA,LAM
    for i in range(N):
        up(param)
        BETA.append(param.beta), LAM.append(param.lam)
        if (n_active > 0  and param.number_act >= n_active) or param.lam == lamin : 
            return (BETA, LAM)

    print('no conv')
    return (BETA, LAM)

def solve_path_Conc(matrices, stop, n_active=False, lassopath=True):
    (A, C, y) = matrices
    n, d, k = len(A), len(A[0]), len(C)
    # to compute r = (A beta - y)/||y|| more efficientely ; and we set reduclam=lam/stop to 2 so that if stop = 0, the condition reduclam < ||r|| is never furfilled
    A_over_NORMy, y_over_NORMy, reduclam = A / (LA.norm(y)), y / (LA.norm(y)), 2.
    if (lassopath):
        lamin, R = 0, [-y_over_NORMy]
    else:
        lamin, beta_old, reduclam_old, r_old = 0, np.zeros(d), 1., -y_over_NORMy

    param = parameters_for_update(matrices, lamin, 0, "Conc")
    BETA, LAM = [param.beta], [param.lam]
    for i in range(N):

        up(param)
        BETA.append(param.beta), LAM.append(param.lam)
        param.r = A_over_NORMy.dot(param.beta) - y_over_NORMy
        if (stop != 0): reduclam = param.lam / stop
        if (lassopath):
            R.append(param.r)
            if (reduclam <= LA.norm(param.r)) or (param.number_act >= n - k) or (
                    n_active > 0 and param.number_act >= n_active): return (BETA, LAM, R)
        else:
            if reduclam <= LA.norm(param.r): return ((beta_old, param.beta), (reduclam_old, reduclam), (r_old, param.r))
            beta_old, reduclam_old, r_old = param.beta, reduclam, param.r

    print('no conv')
    return (BETA, LAM,R)




# Functions to interpolate the solution path between the breaking points
def pathalgo_general(matrix,path,typ,n_active=False,rho=0):
    BETA, i = [], 0
    X,sp_path = solve_path(matrix,path[-1],n_active,rho,typ)
    sp_path.append(path[-1]),X.append(X[-1])
    for lam in path:
        while (lam<sp_path[i+1]): i+=1
        teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
        BETA.append(X[i]*(1-teta)+X[i+1]*teta)
    return(BETA)

def pathalgo_huber_cl(matrix, path, rho, n_active=False):
    return( pathalgo_general(matrix, path, 'huber_cl', n_active=n_active, rho=rho) )

def pathalgo_cl(matrix,path,n_active=False):
    return(pathalgo_general(matrix,path,'cl',n_active))


def up(param):
    formulation = param.formulation
    if (formulation in ['LS','Conc']  ):
        up_LS(param)
    elif (formulation == 'huber'):
        up_huber(param)
    elif (formulation == 'cl'):
        up_cl(param)
    elif (formulation == 'huber_cl'):
        up_huber_cl(param)


# function that search the next lambda where something happen, and update the solution Beta
def up_LS(param):
    lambdamax, lamin, M, C, eps_L2 = param.lambdamax, param.lamin, param.M, param.C, param.eps_L2
    number_act, idr, Xt, activity, beta, s, lam = param.number_act, param.idr, param.Xt, param.activity, param.beta, param.s, param.lam

    d = len(activity)
    L = [lam] * d
    D, E = direction(activity, s, M[:len(activity), :][:, :len(activity)], M[d:, :][:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], D[i], E[i], s[i]
        if (activity[i]):
            if (abs(bi * di) > 1e-10 and bi * di < 0):
                L[i] = -bi / (di * lambdamax)
        else:
            if (abs(e - s0) < 1e-10): continue
            if (e > s0):
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam
    dlamb = min(min(L), lam - lamin)
    # Update matrix inverse, list of rows in C and activity
    for i in range(d):
        if (L[i] < dlamb + 1e-10):
            if (activity[i]):
                activity[i], number_act = False, number_act - 1
                if (len(M) > d):
                    to_ad = next_idr2(idr, C[:, activity])
                    if (type(to_ad) == int): idr[to_ad] = False
            else:
                #x = M[:, activity + idr][i]
                #al = M[i, i] - np.vdot(x, Xt.dot(x))
                #if (abs(al) < 1e-10): break
                activity[i], number_act = True, number_act + 1
                if (len(M) > d):
                    to_ad = next_idr1(idr, C[:, activity])
                    if (type(to_ad) == int): idr[to_ad] = True
        
            
    try : 
        Xt = LA.inv(M[activity + idr, :][:, activity + idr])
    except LA.LinAlgError :
        Xt = LA.inv( M[activity + idr, :][:, activity + idr] + np.diag([eps_L2]*sum(activity) + [0]*sum(idr) )   )
            

    beta = beta + lambdamax * D * dlamb
    if not (lam == dlamb): s = E + lam / (lam - dlamb) * (s - E)
    lam -= dlamb

    param.number_act, param.idr, param.Xt, param.activity, param.beta, param.s, param.lam = number_act, idr, Xt, activity, beta, s, lam



def up_huber(param):
    lambdamax, lamin, A, y, C, rho, eps_L2 = param.lambdamax, param.lamin, param.A, param.y, param.C, param.rho, param.eps_L2
    number_act, idr, Xt, activity, F, beta, s, lam, M, r = param.number_act, param.idr, param.Xt, param.activity, param.F, param.beta, param.s, param.lam, param.M, param.r
    d = len(activity)
    L = [lam] * d
    D, E = direction(activity, s, M[:len(activity), :][:, :len(activity)], M[d:, :][:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], D[i], E[i], s[i]
        if (activity[i]):
            if (abs(bi * di) > 1e-10 and bi * di < 0):
                L[i] = -bi / (di * lambdamax)
        else:
            if (abs(e - s0) < 1e-10 or abs(s0) > 1): continue
            if (e > s0):
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    huber_up, ADl = False, A.dot(D) * lambdamax
    j_switch = None
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and |r[j]+ADl[j]*dl|>rho
        # or  find if there is a  0< dl < dlamb such that not F[j] |r[j]+ADl[j]*dl|<rho
        if (abs(ADl[j]) < 1e-6 or abs(r[j]) < 1e-6 or abs(abs(r[j]) - rho) < 1e-6): continue
        c = -r[j] / ADl[j]
        dl = (1 - rho / abs(r[j])) * c
        if (dl < 0):
            if (c > 0):
                dl = (1 + rho / abs(r[j])) * c
            else:
                dl = dlamb
        if (dl < dlamb):
            huber_up, j_switch, dlamb = True, j, dl
    beta, s, r, lam = beta + lambdamax * D * dlamb, E + lam / (lam - dlamb) * (s - E), r + ADl * dlamb, lam - dlamb

    if (huber_up):
        F[j_switch] = not F[j_switch]
        M[:d, :][:, :d] = 2 * A[F].T.dot(A[F])
    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if (L[i] < dlamb + 1e-10):
                if (activity[i]):
                    activity[i], number_act = False, number_act - 1
                    if (len(M) > d):
                        to_ad = next_idr2(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = False
                else:
                    activity[i], number_act = True, number_act + 1
                    if (len(M) > d):
                        to_ad = next_idr1(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = True

    try : 
        Xt = LA.inv(M[activity + idr, :][:, activity + idr])
    except LA.LinAlgError :
        Xt = LA.inv( M[activity + idr, :][:, activity + idr] + np.diag([eps_L2]*sum(activity) + [0]*sum(idr) )   )


    param.number_act, param.idr, param.Xt, param.activity, param.F, param.beta, param.s, param.lam, param.M, param.r = number_act, idr, Xt, activity, F, beta, s, lam, M, r

def up_cl(param):
    lambdamax,lamin,A, y, eps_L2  = param.lambdamax,param.lamin,param.A, param.y, param.eps_L2
    number_act,idr,Xt,activity,F,beta,s, lam, M,r =\
        param.number_act,param.idr,param.Xt,param.activity,param.F,param.beta,param.s, param.lam, param.M,param.r

    d = len(activity)
    L = [lam] * d
    D, E = direction(activity, s, M[:len(activity), :][:, :len(activity)], M[d:, :][:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], D[i], E[i], s[i]
        if (activity[i]):
            if (abs(bi * di) > 1e-10 and bi * di < 0):
                L[i] = -bi / (di * lambdamax)
        else:
            if (abs(e - s0) < 1e-10 or abs(s0) > 1): continue
            if (e > s0):
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    max_up, yADl = False, y*A.dot(D) * lambdamax
    j_switch = None
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and r[j]+yADl[j]*dl > 1
        # or  find if there is a  0< dl < dlamb such that not F[j] r[j]+yADl[j]*dl < 1
        if (abs(r[j]-1)<1e-4): continue
        if (yADl[j] != 0.): dl = (1-r[j]) / yADl[j]
        else : dl = -1
        if (dl < dlamb and dl >0):
            max_up, j_switch, dlamb = True, j, dl

    beta, s, r, lam = beta + lambdamax * D * dlamb, E + lam / (lam - dlamb) * (s - E), r + yADl * dlamb, lam - dlamb

    if (max_up):
        F[j_switch] = not F[j_switch]
        M[:d, :][:, :d] = 2 * A[F].T.dot(A[F])
    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if (L[i] < dlamb + 1e-10):
                if (activity[i]):
                    activity[i], number_act = False, number_act - 1
                    if (len(M) > d):
                        to_ad = next_idr2(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = False
                else:
                    # x = M[:,activity+idr][i]
                    # al = M[i,i]-np.vdot(x,Xt.dot(x))
                    # if (abs(al)<1e-10): break
                    activity[i], number_act = True, number_act + 1
                    if (len(M) > d):
                        to_ad = next_idr1(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = True
    try : 
        Xt = LA.inv(M[activity + idr, :][:, activity + idr])
    except LA.LinAlgError :
        Xt = LA.inv( M[activity + idr, :][:, activity + idr] + np.diag([eps_L2]*sum(activity) + [0]*sum(idr) )   )

    param.number_act, param.idr, param.Xt, param.activity, param.F, param.beta, param.s, param.lam, param.M, param.r = \
        number_act, idr, Xt, activity, F, beta, s, lam, M, r

def up_huber_cl(param):
    lambdamax, lamin, A, y, rho, eps_L2 = param.lambdamax, param.lamin, param.A, param.y, param.rho, param.eps_L2
    number_act, idr, Xt, activity, F, beta, s, lam, M, r = param.number_act, param.idr, param.Xt, param.activity, param.F, param.beta, param.s, param.lam, param.M, param.r
    d = len(activity)
    L = [lam] * d
    D, E = direction(activity, s, M[:len(activity), :][:, :len(activity)], M[d:, :][:, :d], Xt, idr, number_act)
    for i in range(d):
        bi, di, e, s0 = beta[i], D[i], E[i], s[i]
        if (activity[i]):
            if (abs(bi * di) > 1e-10 and bi * di < 0):
                L[i] = -bi / (di * lambdamax)
        else:
            if (abs(e - s0) < 1e-10 or abs(s0) > 1): continue
            if (e > s0):
                dl = (1 + s0) / (1 + e)
            else:
                dl = (1 - s0) / (1 - e)
            L[i] = dl * lam

    dlamb = min(min(L), lam, lam - lamin)
    max_up, yADl = False, y * A.dot(D) * lambdamax
    j_switch = None
    for j in range(len(r)):
        #     find if there is a  0< dl < dlamb such that F[j] and r[j]+yADl[j]*dl > 1
        # or  find if there is a  0< dl < dlamb such that not F[j] r[j]+yADl[j]*dl < 1
        #     find if there is a  0< dlhuber < dlamb such that F[j] and r[j]+yADl[j]*dl < rho
        # or  find if there is a  0< dlhuber < dlamb such that not F[j] r[j]+yADl[j]*dl > rho


        if (yADl[j] != 0.):
            dlmax      = (1 - r[j]) / yADl[j]
            dlhuber = (rho - r[j]) / yADl[j]
        else: dlmax,dlhuber = dlamb+1,dlamb+1
        if(dlmax<=0.):     dlmax   = dlamb+1
        if(dlhuber <= 0.): dlhuber = dlamb+1

        dl = min ( dlhuber , dlmax)

        if (dl < dlamb) : max_up, j_switch, dlamb = True, j, dl

    beta, s, r, lam = beta + lambdamax * D * dlamb, E + lam / (lam - dlamb) * (s - E), r + yADl * dlamb, lam - dlamb

    if (max_up):
        F[j_switch] = not F[j_switch]
        M[:d, :][:, :d] = 2 * A[F].T.dot(A[F])

    else:
        # Update matrix inverse, list of rows in C and activity
        for i in range(d):
            if (L[i] < dlamb + 1e-10):
                if (activity[i]):
                    activity[i], number_act = False, number_act - 1
                    if (len(M) > d):
                        to_ad = next_idr2(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = False
                else:
                    # x = M[:,activity+idr][i]
                    # al = M[i,i]-np.vdot(x,Xt.dot(x))
                    # if (abs(al)<1e-10): break
                    activity[i], number_act = True, number_act + 1
                    if (len(M) > d):
                        to_ad = next_idr1(idr, M[d:, :][:, :d][:, activity])
                        if (type(to_ad) == int): idr[to_ad] = True

    try : 
        Xt = LA.inv(M[activity + idr, :][:, activity + idr])
    except LA.LinAlgError :
        Xt = LA.inv( M[activity + idr, :][:, activity + idr] + np.diag([eps_L2]*sum(activity) + [0]*sum(idr) )   )

    param.number_act, param.idr, param.Xt, param.activity, param.F, param.beta, param.s, param.lam, param.M, param.r = number_act, idr, Xt, activity, F, beta, s, lam, M, r


















# Compute the derivatives of the solution Beta and the derivative of lambda*subgradient thanks to the ODE
def direction(activity,s,Mat,C,Xt,idr,number_act):
    if (len(C)==0):
        D,product =np.zeros(len(activity)), Xt[:,:number_act].dot(s[activity])
        D[activity]= product
        return(D,Mat.dot(D))
    D,Dadj=np.zeros(len(activity)),np.zeros(len(C))
    product = Xt[:,:number_act].dot(s[activity])
    D[activity],Dadj[idr]=product[:number_act],product[number_act:]
    E = (Mat.dot(D)+C.T.dot(Dadj))   #D and D2 in Rd with zeros in inactives and E. D is - derivatives
    return(D,E)

#Upddate a list of constraints which are independant if we restrict the matrix C to the acrive set (C_A has to have independant rows)
#When we ad an active parameter
def next_idr1(liste,mat):
    if(sum(liste)==len(mat)): return(False)
    if (sum(liste)==0):
        for i in range(len(mat)):
            for j in range(len(mat[0])):
                if not (mat[i,j]==0): return(i)
        return(False)
    Q = LA.qr(mat[liste,:].T)[0]
    for j in range(len(mat)):
        if (not liste[j]) and (  LA.norm(mat[j]-LA.multi_dot([Q,Q.T,mat[j]]))>1e-10 ): return(j)
    return(False)

# When we remove an active parameter
def next_idr2(liste, mat):
    if (sum(liste) == 0): return (False)
    R = LA.qr(mat[liste, :].T)[1]
    for i in range(len(R)):
        if (abs(R[i, i]) < 1e-10):  # looking for the i-th True element of liste
            j, somme = 0, liste[0]
            while (somme <= i):
                j, somme = j + 1, somme + liste[j + 1]
            return (j)
    return (False)

# Update the invers of a matrix whom we add a line, which is useul to compute the derivatives
def next_inv(Xt, B, al, ligne):
    n = len(Xt)
    Yt = np.zeros((n + 1, n + 1))
    alpha = 1 / al
    B = np.array([B])
    b1 = Xt[:ligne, :][:, :ligne] + alpha * B[:, :ligne].T.dot(B[:, :ligne])
    b2 = Xt[ligne:, :][:, :ligne] + alpha * B[:, ligne:].T.dot(B[:, :ligne])
    b4 = Xt[ligne:, :][:, ligne:] + alpha * B[:, ligne:].T.dot(B[:, ligne:])
    col1 = np.concatenate((b1, -alpha * B[:, :ligne], b2), axis=0)
    col2 = np.concatenate((b2.T, -alpha * B[:, ligne:], b4), axis=0)
    col = np.concatenate((-alpha * B[0, :ligne], [alpha], -alpha * B[0, ligne:]), axis=0)
    return (np.concatenate((col1, np.array([col]).T, col2), axis=1))

def h_lambdamax(X, y, rho):
    return 2 * LA.norm(X.T.dot(h_prime(y, rho)), np.infty)

# Compute the derivative of the huber function, particulary useful for the computing of lambdamax
def h_prime(y,rho):
    if (rho==0): return(y)
    m = len(y)
    lrho = rho*np.ones(m)
    return(np.maximum(lrho,-y)+ np.minimum(y-lrho,0))




'''
Function that solve the concomitant problem for every lambda thanks to the previous function : we firstly compute all the non concomitant Least square problems, then we use it to find sigma and so the solution, using the equation for sigma: 

sigma = || A*B(lambda*sigma) - y ||_2       (where B(lambda) is found thanks to solve_path)

            teta = (sp_path[i]-lam)/(sp_path[i]-sp_path[i+1])
'''



import numpy as np
import numpy.random as rd
import numpy.linalg as LA
from .compact_func import Classo,pathlasso
n_lam = 200

def compute_1SE(mse_max,MSE,i): 
    j=i
    while(j> 0 and MSE[j]<mse_max): j-=1
    return j


def train_test_CV(n,k,test_pourcent):
    idx, training_size = rd.permutation(n), int(n-n*test_pourcent)
    idx_train, idx_test = idx[:training_size], idx[training_size:]
    SUBLIST,end = [],0 
    for i in range(k):
        begin,end = end,end+training_size//k
        if(i<training_size%k): end+=1
        SUBLIST.append(idx[begin:end])
    return(SUBLIST, idx_train,idx_test) 


def train_test_i (SUBLIST,i):
    training_set,test_set = np.array([],dtype=int),SUBLIST[i]
    for j in range(len(SUBLIST)):
        if (j != i): training_set = np.concatenate((training_set,SUBLIST[j]))
    return(training_set,test_set)
            

def training(matrices,typ,num_meth, training_set, rho, rho_classification, e,lambdas):
    (A,C,y)   = matrices
    mat       = (A[training_set],C,y[training_set]) 
    sol       = pathlasso(mat,lambdas = lambdas,typ=typ,meth=num_meth,
                            rho = rho, e=e,rho_classification=rho_classification)[0]
    return(sol)


def test_i (matrices,typ,num_meth, SUBLIST,i, rho, rho_classification, e,lambdas):
    training_set,test_set = train_test_i (SUBLIST,i)
    BETA                  = training(matrices,typ,num_meth, training_set, rho, rho_classification, e, lambdas)
    n_lam = len(lambdas)
    residual = np.zeros(n_lam)
    for j in range(n_lam):
        residual[j] = accuracy_func(matrices[0][test_set],matrices[2][test_set],BETA[j],typ)/len(test_set)
    return(residual)

def average_test(matrices,typ,num_meth, SUBLIST, rho, rho_classification, e,lambdas):
    k = len(SUBLIST)
    n_lam = len(lambdas)
    RESIDUAL = np.zeros((k,n_lam))
    for i in range(k):
        RESIDUAL[i,:] = test_i (matrices,typ,num_meth, SUBLIST,i, rho, rho_classification, e, lambdas)
    MSE = np.mean(RESIDUAL,axis = 0)
    SE = np.std(RESIDUAL,axis = 0) # official standard error should be divided by sqrt(k) ... 
    return(MSE,SE)

def CV(matrices,k,typ='LS',num_meth="Path-Alg",test=0., seed = 1, rho = 1.345, rho_classification=-1., e= 1.,lambdas = np.linspace(1.,1e-3,n_lam),oneSE = True):
    
    rd.seed(seed)
    (A,C,y) = matrices
    SUBLIST, idx_train, idx_test = train_test_CV(len(y),k,test)
    MSE,SE  = average_test(matrices,typ,num_meth, SUBLIST, rho, rho_classification, e, lambdas)
    i       = np.argmin(MSE)
    i_1SE   = compute_1SE(MSE[i]+SE[i],MSE,i)
    if oneSE : lam = lambdas[i_1SE]
    else     : lam = lambdas[i]
    out = Classo((A[idx_train],C,y[idx_train]),lam,typ=typ,meth=num_meth,rho=rho, e=e,rho_classification=rho_classification)
    return(out,MSE,SE,i,i_1SE)
    
        
# Computation of the residual, for huber, LS, huber_classification and classification
def hub(r,rho) : 
    h=0
    for j in range(len(r)):
        if(abs(r[j])<rho): h+=r[j]**2
        elif(r[j]>0)     : h+= (2*r[j]-rho)*rho
        else             : h+= (-2*r[j]-rho)*rho
    return(h)


def hinge(A,y,beta):
    return(sum(np.max(0,1 - y * A.dot(beta))**2))

def huber_hinge(A,y,beta, rho):
    h = np.max(0,1 - y * A.dot(beta))
    s=0
    for i in range(len(h)):
        if h[i]<rho : s+=2*h[i]*rho - rho**2
        else :        s+=h[i]**2
    return s


def accuracy_func(A,y,beta, typ='LS',rho = 1.345, rho_classification=-1.):
    if (typ == 'Huber'):                   return(hub( A.dot(beta) - y , rho ) )
    elif (typ == 'Classification') :       return(hinge(A,y,beta))
    elif (typ == 'Huber_Classification') : return(huber_hinge(A,y,beta,rho_classification))
    else :                                 return(LA.norm( A.dot(beta) - y )**2)
